/*! Tripetto Runner Foundation 6.0.5-1 - Copyright (C) 2021 Tripetto B.V. - All Rights Reserved */

/// <reference path="marblecore.matrix.d.ts" />
/// <reference path="marblecore.callback.d.ts" />
/// <reference path="tripetto.slots.d.ts" />
/// <reference path="marblecore.serializer.d.ts" />
/// <reference path="marblecore.hookup.d.ts" />
/// <reference path="marblecore.l10n.d.ts" />
/// <reference path="marblecore.functional.d.ts" />

declare module "@tripetto/vfsm" {
    /** Explicit exports (fixed order) */
    export { Conditions } from "@tripetto/vfsm/lib/conditions/conditions";
    export { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    export { Branches } from "@tripetto/vfsm/lib/branches/branches";
    export { Branch } from "@tripetto/vfsm/lib/branches/branch";
    export { States } from "@tripetto/vfsm/lib/states/states";
    export { State } from "@tripetto/vfsm/lib/states/state";
    export { Transducers } from "@tripetto/vfsm/lib/transducers/transducers";
    export { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    /** Interfaces */
    export { IBranch } from "@tripetto/vfsm/lib/branches/interface";
    export { IChainLookup } from "@tripetto/vfsm/lib/chain";
    export { ICondition } from "@tripetto/vfsm/lib/conditions/interface";
    export {
        IDataValue,
        IDataValues,
        IDataImport,
    } from "@tripetto/vfsm/lib/instance/data/interface";
    export { IMetrics } from "@tripetto/vfsm/lib/metrics";
    export { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    export { IMoment } from "@tripetto/vfsm/lib/instance/moments/moment";
    export { Instance } from "@tripetto/vfsm/lib/instance/instance";
    export { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    export { IInstanceSnapshot } from "@tripetto/vfsm/lib/instance/snapshot";
    export { ISnapshot } from "@tripetto/vfsm/lib/snapshot";
    export { IState } from "@tripetto/vfsm/lib/states/interface";
    export { ITransducer } from "@tripetto/vfsm/lib/transducers/interface";
    export { IVariable } from "@tripetto/vfsm/lib/instance/data/variable";
    /** Other explicit exports */
    export { Context } from "@tripetto/vfsm/lib/instance/data/context";
    export { Data } from "@tripetto/vfsm/lib/instance/data/data";
    export { Ontology, OntologyProcessor } from "@tripetto/vfsm/lib/ontology";
    export { Value } from "@tripetto/vfsm/lib/instance/data/value";
    export { Values } from "@tripetto/vfsm/lib/instance/data/values";
    /** Enumerations */
    export { Errors } from "@tripetto/vfsm/enum/errors";
    /** Decorators */
    export { ontology } from "@tripetto/vfsm/lib/decorator";
    /** Events */
    export {
        IOntologyEndEvent,
        IOntologyFinishEvent,
        IOntologyInstanceCreateEvent,
        IOntologyInstanceEndEvent,
        IOntologyInstancePauseEvent,
        IOntologyInstanceStartEvent,
        IOntologyInstanceProcessingEvent,
        IOntologyInstanceStopEvent,
        IOntologyLoadEvent,
        IOntologyPauseEvent,
        IOntologyResumeEvent,
        IOntologyStartEvent,
        IOntologyStopEvent,
        IOntologyUnloadEvent,
    } from "@tripetto/vfsm/lib/events";
    export {
        IInstanceCreateEvent,
        IInstanceEndEvent,
        IInstancePauseEvent,
        IInstanceStartEvent,
        IInstanceStopEvent,
    } from "@tripetto/vfsm/lib/instance/events";
    export {
        IBranchCancelEvent,
        IBranchEnterEvent,
        IBranchLeaveEvent,
        IBranchReadyEvent,
        IBranchSkipEvent,
    } from "@tripetto/vfsm/lib/branches/events";
    export {
        IStateCancelEvent,
        IStateEnterEvent,
        IStateEvaluateEvent,
        IStateLeaveEvent,
        IStateSkipEvent,
        IStateValidatedEvent,
    } from "@tripetto/vfsm/lib/states/events";
    export {
        ITransducerCancelEvent,
        ITransducerEnterEvent,
        ITransducerLeaveEvent,
        ITransducerSkipEvent,
        ITransducerValidatedEvent,
    } from "@tripetto/vfsm/lib/transducers/events";
    export {
        IValueChangeEvent,
        ISealChangeEvent,
        IDataChangeEvent,
    } from "@tripetto/vfsm/lib/instance/data/events";
}

declare module "@tripetto/vfsm/lib/conditions/conditions" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { ICondition } from "@tripetto/vfsm/lib/conditions/interface";
    import { StrictMatrix } from "@marblecore/matrix";
    export class Conditions extends StrictMatrix<
        Conditions,
        Condition,
        void,
        ICondition
    > {
        /**
         * Constructs a new condition dispatcher.
         * @param Reference to the parent branch.
         */
        constructor(pBranch: Branch);
        /** Retrieves a reference to the parent branch. */
        get branch(): Branch;
        /**
         * Creates a condition.
         * @param pCondition Reference to the condition ontology.
         * @return Returns a reference to the condition instance or `undefined` if
         * a similar condition was already found.
         */
        create<T extends Condition>(pCondition: ICondition): T | undefined;
    }
}

declare module "@tripetto/vfsm/lib/conditions/condition" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Callback } from "@marblecore/callback";
    import { Conditions } from "@tripetto/vfsm/lib/conditions/conditions";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { Factory } from "@tripetto/vfsm/lib/factory";
    import { ICondition } from "@tripetto/vfsm/lib/conditions/interface";
    import { State } from "@tripetto/vfsm/lib/states/state";
    export abstract class Condition<
        T extends ICondition = ICondition
    > extends Factory<Conditions, T> {
        /**
         * Creates the context for the given condition props.
         * @param pCondition Reference to the condition props.
         * @return Returns the context.
         */
        static createContext<T extends ICondition>(pCondition: T): string;
        /**
         * Constructs a new condition.
         * @param pConditions Reference to the condition dispatcher.
         * @param pCondition Reference to the condition props.
         */
        constructor(pConditions: Conditions, pCondition: T);
        /** Reference to the conditions dispatcher. */
        protected get conditions(): Conditions;
        /** Reference to the parent branch. */
        get branch(): Branch;
        /** Reference to the parent state. */
        get state(): State;
        /** Retrieves the condition context. */
        get context(): string;
        /** Retrieves the condition properties. */
        get props(): Readonly<T>;
        /**
         * Invoked when the VFSM wants to verify the condition.
         * @param pContext Reference to the context.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous condition callback.
         * @return Returns the result of the condition or the callback reference if
         * the condition should be handled asynchronously.
         * @event
         */
        abstract onCondition(
            pContext: Context,
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when the condition is approved.
         * @param pContext Reference to the context.
         * @event
         */
        onApproved(pContext: Context): void;
    }
}

declare module "@tripetto/vfsm/lib/branches/branches" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { IBranch } from "@tripetto/vfsm/lib/branches/interface";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { StrictMatrix } from "@marblecore/matrix";
    export class Branches extends StrictMatrix<
        Branches,
        Branch,
        void,
        IBranch
    > {
        /**
         * Constructs a new branch dispatcher.
         * @param pState Reference to the parent state.
         */
        constructor(pState: State);
        /** Retrieves a reference to the parent state. */
        get state(): State;
        /**
         * Creates a branch.
         * @param pBranch Reference to the branch ontology.
         * @param pAncestor Reference to the ancestor state.
         * @return Returns a reference to the branch instance.
         */
        create<T extends Branch>(pBranch: IBranch, pAncestor: State): T;
    }
}

declare module "@tripetto/vfsm/lib/branches/branch" {
    /** Dependencies */
    import { Branches } from "@tripetto/vfsm/lib/branches/branches";
    import { Callback } from "@marblecore/callback";
    import { Conditions } from "@tripetto/vfsm/lib/conditions/conditions";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { Factory } from "@tripetto/vfsm/lib/factory";
    import { IBranch } from "@tripetto/vfsm/lib/branches/interface";
    import { TBranchEvents } from "@tripetto/vfsm/lib/branches/events";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { TBranchCulling } from "@tripetto/vfsm/lib/branches/culling";
    export abstract class Branch<T extends IBranch = IBranch> extends Factory<
        Branches,
        T,
        TBranchEvents
    > {
        /**
         * Constructs a new branch.
         * @param pBranches Reference to the branch dispatcher.
         * @param pBranch Reference to the branch data.
         */
        constructor(pBranches: Branches, pBranch: T);
        /** Retrieves a reference to the branches. */
        protected get branches(): Branches;
        /** Retrieves an array with states. */
        get states(): State[];
        /** Retrieves a reference to the parent state. */
        get state(): State;
        /** Retrieves a reference to the condition dispatcher. */
        get conditions(): Conditions;
        /** Retrieves the branch properties. */
        get props(): Readonly<T>;
        /** Retrieves the culling mode for the branch. */
        get culling(): TBranchCulling;
        /** Retrieves if this branch should spawn a new stack. */
        get spawn(): boolean;
        /** Retrieves if this is the main branch. */
        get isMain(): boolean;
        /** Retrieves if this branch is a finish. */
        get isFinish(): boolean;
        /**
         * Retrieves the entry state for the branch.
         * @param sMode Specifies the mode.
         * @return Returns the entry state.
         */
        entry(sMode: "live" | "preview"): State | undefined;
        /**
         * Invoked when the VFSM wants to verify the condition for the branch.
         * @param pContext Reference to the context.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous condition callback.
         * @return Returns the result of the condition or the callback reference if
         * the condition should be handled asynchronously.
         * @event
         */
        abstract onCondition(
            pContext: Context,
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when the branch is ready for entering. The condition is approved.
         * @param pContext Reference to the context.
         * @event
         */
        onReady(pContext: Context): void;
        /**
         * Invoked when the branch is entered.
         * @param pContext Reference to the context.
         * @event
         */
        onEnter(pContext: Context): void;
        /**
         * Invoked when the branch is skipped due to a failing enter condition.
         * @param pContext Reference to the context.
         * @event
         */
        onSkip(pContext: Context): void;
        /**
         * Invoked when the branch is left.
         * @param pContext Reference to the context.
         * @event
         */
        onLeave(pContext: Context): void;
        /**
         * Invoked when the branch is canceled.
         * @param pContext Reference to the context.
         * @event
         */
        onCancel(pContext: Context): void;
    }
}

declare module "@tripetto/vfsm/lib/states/states" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { IState } from "@tripetto/vfsm/lib/states/interface";
    import { Main } from "@tripetto/vfsm/lib/states/main";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { StrictMatrix } from "@marblecore/matrix";
    import { TOntology } from "@tripetto/vfsm/lib/ontology";
    export class States extends StrictMatrix<States, State, void, IState> {
        /**
         * Constructs a new state dispatcher.
         * @param pOntology Reference to the ontology.
         */
        constructor(pOntology: TOntology);
        /** Retrieves a reference to the ontology. */
        get ontology(): TOntology;
        /**
         * Creates a new main state.
         * @return Returns the main state.
         */
        main(): Main;
        /**
         * Creates a state.
         * @param pState Reference to the state data.
         * @param pBranch Reference to the parent branch.
         * @param pAncestor Reference to the ancestor state.
         * @param nEnumerator Specifies the enumerator of the state.
         * @return Returns a reference to the state.
         */
        create<T extends State>(
            pState: IState,
            pBranch: Branch,
            pAncestor: State,
            nEnumerator: number
        ): T;
    }
}

declare module "@tripetto/vfsm/lib/states/state" {
    /** Dependencies */
    import { Await, Callback, ICallbackPromise } from "@marblecore/callback";
    import { Factory } from "@tripetto/vfsm/lib/factory";
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Branches } from "@tripetto/vfsm/lib/branches/branches";
    import { Transducers } from "@tripetto/vfsm/lib/transducers/transducers";
    import { States } from "@tripetto/vfsm/lib/states/states";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { TOntology } from "@tripetto/vfsm/lib/ontology";
    import { IState } from "@tripetto/vfsm/lib/states/interface";
    import { TStateEvents } from "@tripetto/vfsm/lib/states/events";
    export abstract class State<T extends IState = IState> extends Factory<
        States,
        T,
        TStateEvents
    > {
        /** Contains the await pointer for the state. */
        await: Await | undefined;
        /**
         * Constructs a new state.
         * @param pStates Reference to the state dispatcher.
         * @param pState Specifies the state data.
         */
        constructor(pStates: States, pState: T);
        /** Retrieves a reference to the states. */
        get states(): States;
        /** Retrieves a reference to the ontology. */
        get ontology(): TOntology;
        /** Retrieves a reference to the transducer dispatcher. */
        get transducers(): Transducers;
        /** Retrieves a reference to the branch dispatcher. */
        get branches(): Branches;
        /** Retrieves if the state is the main state. */
        get isMain(): boolean;
        /** Retrieves the state properties. */
        get props(): Readonly<T>;
        /** Retrieves if the state is evaluating. */
        get isEvaluating(): boolean;
        /**
         * Contains the state enumerator. This is a chronological number for each
         * state. The first state has a enumerator value of `1` and the last state
         * equals the value of the `metrics.states` of a ontology instance. You can
         * use this property to calculate a simple progress indicator.
         */
        get enumerator(): number;
        /**
         * Prepares the instance.
         * @param pAncestor Reference to the ancestor.
         * @param pBranch Reference to the branch.
         * @param nEnumerator Specifies the enumerator value.
         */
        prepare(
            pAncestor: State,
            pBranch: Branch,
            nEnumerator?: number
        ): State<T>;
        /**
         * Evaluates the state.
         * @param pReference Reference to the callback or promise.
         * @param pContext Reference to the context.
         */
        evaluate<C>(
            pReference: Callback<C> | ICallbackPromise<C>,
            pContext: Context,
            fnAssert: (pPromise: ICallbackPromise<C>) => boolean
        ): void;
        /**
         * Invoked when the VFSM wants to verify the enter condition for the state.
         * This event is invoked just before the state is entered and it is also
         * invoked for the main state. This allows you to use a condition for the
         * main state which lets you control the main application branch. If you
         * override this function you have to make sure to return an appropriate
         * value for the main state.
         * @param pContext Reference to the context.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous condition callback.
         * @return Returns the result of the condition or the callback reference if
         * the condition should be handled asynchronously.
         * @event
         */
        abstract onEnterCondition(
            pContext: Context,
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when state is entered.
         * @param pContext Reference to the context.
         * @param pAwait Reference to the await pointer. Pass this pointer as return
         * value to enable the asynchronous callback.
         * @return Returns the callback reference if the function should be handled
         * asynchronously.
         * @event
         */
        abstract onEnter(pContext: Context, pAwait: Await): void | Await;
        /**
         * Invoked when state is skipped due to a failing enter condition.
         * @param pContext Reference to the context.
         * @event
         */
        onSkip(pContext: Context): void;
        /**
         * Invoked when the VFSM wants to verify the leave condition for the state.
         * This event is invoked just before the state is left.
         * @param pContext Reference to the context.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous condition callback.
         * @return Returns the result of the condition or the callback reference if
         * the condition should be handled asynchronously.
         * @event
         */
        abstract onLeaveCondition(
            pContext: Context,
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when the state is left.
         * @param pContext Reference to the context.
         * @return Returns the callback reference if the function should be handled
         * asynchronously.
         * @event
         */
        onLeave(pContext: Context): void;
        /**
         * Invoked when the state cannot be left due to a failing leave condition.
         * @param pContext Reference to the context.
         * @param pAwait Reference to the await pointer. Pass this pointer as return
         * value to enable the asynchronous callback.
         * @event
         */
        onLeaveError(pContext: Context, pAwait: Await): void | Await;
        /**
         * Invoked when the state is canceled.
         * @param pContext Reference to the context.
         * @event
         */
        onCancel(pContext: Context): void;
        /**
         * Invoked when the state needs to be validated.
         * @param pContext Reference to the context.
         * @param sCurrent Specifies the current validation state.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous validation callback.
         * @return Returns the result of the validation or the callback reference if
         * the validation should be handled asynchronously.
         * @event
         */
        abstract onValidate(
            pContext: Context,
            sCurrent: "unknown" | "fail" | "pass",
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when the state is validated.
         * @param pContext Reference to the context.
         * @param sResult Specifies the validation state.
         * @param sType Specifies if the validation type.
         * @event
         */
        onValidated(
            pContext: Context,
            sResult: "fail" | "pass",
            sType: "initial" | "revalidate"
        ): void;
        /**
         * Invoked when the state starts or stops evaluating.
         * @param pContext Reference to the context.
         * @param bEvaluating Specifies if the state is evaluating.
         * @event
         */
        onEvaluate(pContext: Context, bEvaluating: boolean): void;
    }
}

declare module "@tripetto/vfsm/lib/transducers/transducers" {
    /** Dependencies */
    import { StrictMatrix } from "@marblecore/matrix";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { ITransducer } from "@tripetto/vfsm/lib/transducers/interface";
    export class Transducers extends StrictMatrix<
        Transducers,
        Transducer,
        void,
        ITransducer
    > {
        /**
         * Constructs a new transducer dispatcher.
         * @param pState Reference to the state.
         */
        constructor(pState: State);
        /** Retrieves a reference to the parent state. */
        get state(): State;
        /**
         * Creates a transducer.
         * @param pTransducer Reference to the transducer ontology.
         * @return Returns a reference to the transducer instance.
         */
        create<T extends Transducer>(pTransducer: ITransducer): T;
    }
}

declare module "@tripetto/vfsm/lib/transducers/transducer" {
    /** Dependencies */
    import { Await, Callback } from "@marblecore/callback";
    import { Slots } from "@tripetto/slots";
    import { Factory } from "@tripetto/vfsm/lib/factory";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducers } from "@tripetto/vfsm/lib/transducers/transducers";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { ITransducer } from "@tripetto/vfsm/lib/transducers/interface";
    import { TTransducerEvents } from "@tripetto/vfsm/lib/transducers/events";
    export abstract class Transducer<
        T extends ITransducer = ITransducer
    > extends Factory<Transducers, T, TTransducerEvents> {
        /** Contains the await pointer for the transducer. */
        await: Await | undefined;
        /**
         * Constructs a new transducer.
         * @param pTransducers Reference to the transducer dispatcher.
         * @param pTransducer Reference to the transducer data.
         */
        constructor(pTransducers: Transducers, pTransducer: T);
        /** Retrieves a reference to the transducers. */
        protected get transducers(): Transducers;
        /** Retrieves a reference to the parent state. */
        get state(): State;
        /** Retrieves a reference to the slots. */
        get slots(): Slots;
        /** Retrieves the transducer properties. */
        get props(): Readonly<T>;
        /**
         * Seal or unseal the values.
         * @param pContext Reference to the context.
         * @param bSeal Specifies if the values should be sealed.
         */
        seal(pContext: Context, bSeal: boolean): void;
        /**
         * Retrieves if the seal should be verified.
         * @param pContext Reference to the context.
         * @return Returns `true` if the seal needs verification.
         */
        shouldVerifySeal(pContext: Context): boolean;
        /**
         * Retrieves if one or more of the transducer slots are awaiting a value.
         * @param pContext Reference to the context.
         * @return Returns `true` if the seal needs verification.
         */
        isAwaiting(pContext: Context): boolean;
        /**
         * Invoked when the VFSM wants to verify the condition for the transducer.
         * This event is invoked just before the transducer is entered.
         * @param pContext Reference to the context.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous condition callback.
         * @return Returns the result of the condition or the callback reference if
         * the condition should be handled asynchronously.
         * @event
         */
        abstract onEnterCondition(
            pContext: Context,
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when the transducer is entered.
         * @param pContext Reference to the context.
         * @param pAwait Reference to the await pointer. Pass this pointer as return
         * value to enable asynchronous validation.
         * @return Return the await pointer to implement asynchronous transducer validation.
         * @event
         */
        abstract onEnter(pContext: Context, pAwait: Await): void | Await;
        /**
         * Invoked when the transducer is skipped due to a failing enter condition.
         * @param pContext Reference to the context.
         * @event
         */
        onSkip(pContext: Context): void;
        /**
         * Invoked when the transducer is left.
         * @param pContext Reference to the context.
         * @event
         */
        onLeave(pContext: Context): void;
        /**
         * Invoked when the transducer is canceled.
         * @param pContext Reference to the context.
         * @event
         */
        onCancel(pContext: Context): void;
        /**
         * Invoked before the transducer is validated. Can be used to do data
         * sanitizing before validation occurs.
         * @param pContext Reference to the context.
         * @event
         */
        onBeforeValidate(pContext: Context): void;
        /**
         * Invoked when the transducer is validated.
         * @param pContext Reference to the context.
         * @param sCurrent Specifies the current transducer validation state.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous validation callback.
         * @return Returns the result of the validation or the callback reference if
         * the validation should be handled asynchronously.
         * @event
         */
        abstract onValidate(
            pContext: Context,
            sCurrent: "unknown" | "fail" | "pass",
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when the transducer is validated.
         * @param pContext Reference to the context.
         * @param sResult Specifies the validation result.
         * @param sType Specifies the validation type.
         * @event
         */
        onValidated(
            pContext: Context,
            sResult: "fail" | "pass",
            sType: "initial" | "revalidate"
        ): void;
        /**
         * Invoked when a side effect is generated.
         * @param pContext Reference to the context.
         * @param pTransducer Specifies the transducer that caused the side effect.
         * @event
         */
        onSideEffect(pContext: Context, pTransducer: Transducer): void;
    }
}

declare module "@tripetto/vfsm/lib/branches/interface" {
    /** Dependencies */
    import { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    import { TBranchCulling } from "@tripetto/vfsm/lib/branches/culling";
    export interface IBranch extends IOntologyProperties {
        /**
         * Specifies the jump reference for the branch. If this property is omitted
         * or an empty string, the branch successor will be determined
         * automatically. If this property is set to 'abort' the stack pointer will
         * be killed after the branch. If this property is set to the identifier
         * of another state, the stack pointer is redirected to that state.
         */
        readonly jump?: string | "abort";
        /** Specifies the culling mode for the branch. */
        readonly culling?: TBranchCulling;
        /** Specifies if this branch should spawn a new stack. */
        readonly spawn?: boolean;
    }
}

declare module "@tripetto/vfsm/lib/chain" {
    /** Dependencies */
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { Slot } from "@tripetto/slots";
    export interface IChain {
        readonly states: {
            readonly [hash: string]: State | undefined;
        };
        readonly transducers: {
            readonly [hash: string]: Transducer | undefined;
        };
        readonly branches: {
            readonly [hash: string]: Branch | undefined;
        };
        readonly conditions: {
            readonly [hash: string]: Condition | undefined;
        };
        readonly slots: {
            readonly [hash: string]: Slot[] | Slot | undefined;
        };
    }
    export interface IChainLookup {
        /** Retrieves a state. */
        readonly state: (sHash: string) => State | undefined;
        /** Retrieves a transducer. */
        readonly transducer: (sHash: string) => Transducer | undefined;
        /** Retrieves a branch. */
        readonly branch: (sHash: string) => Branch | undefined;
        /** Retrieves a condition. */
        readonly condition: (sHash: string) => Condition | undefined;
        /** Retrieves a slot or slot pipe. */
        readonly slot: (sHash: string) => Slot | Slot[] | undefined;
    }
}

declare module "@tripetto/vfsm/lib/conditions/interface" {
    /** Dependencies */
    import { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    export interface ICondition extends IOntologyProperties {}
}

declare module "@tripetto/vfsm/lib/instance/data/interface" {
    /** Dependencies */
    import { TSerializeTypes } from "@marblecore/serializer";
    export interface IDataValue {
        /** Contains the value. */
        readonly value: TSerializeTypes;
        /** Contains the reference. */
        readonly reference: string | undefined;
        /** Contains the UTC set time of the value. */
        readonly time: number;
    }
    export interface IDataValues {
        readonly [context: string]: IDataValue | undefined;
    }
    export interface IDataImport {
        readonly [context: string]: Partial<IDataValue> | undefined;
    }
}

declare module "@tripetto/vfsm/lib/metrics" {
    export interface IMetrics {
        /** Number of states. */
        states: number;
        /** Number of transducers. */
        transducers: number;
        /** Number of branches. */
        branches: number;
        /** Number of conditions. */
        conditions: number;
        /** Number of slots. */
        slots: number;
    }
}

declare module "@tripetto/vfsm/lib/instance/data/immutable" {
    /** Dependencies */
    import { Slot } from "@tripetto/slots";
    import { THookContext } from "@marblecore/hookup";
    import { TSerializeTypes } from "@marblecore/serializer";
    import { Value } from "@tripetto/vfsm/lib/instance/data/value";
    export class ImmutableValue<
        T extends TSerializeTypes = TSerializeTypes,
        S extends Slot<T> = Slot<T>
    > {
        /**
         * Creates an immutable value instance.
         * @param pValue Reference to the mutable value instance.
         */
        constructor(pValue: Value<T, S>);
        /** Retrieves a unique key for the value within its context (the key is prefixed with an underscore). */
        get key(): string;
        /** Retrieves the context for the value. */
        get context(): "*" | string;
        /** Retrieves a reference to the slot. */
        get slot(): S;
        /** Retrieves the value. */
        get value(): T;
        /**
         * Retrieves the value as a string (this will always return a string with a
         * value regardless of a pristine value was set).
         */
        get string(): string;
        /** Retrieves the optional value reference. */
        get reference(): string | undefined;
        /** Retrieves the value set timestamp. */
        get time(): number | undefined;
        /** Retrieves if there is a value. */
        get hasValue(): boolean;
        /** Retrieves if the value is modified. */
        get isModified(): boolean;
        /** Retrieves if the value is sealed. */
        get isSealed(): boolean;
        /** Retrieves if the value is locked. */
        get isLocked(): boolean;
        /** Retrieves if the value is frozen. */
        get isFrozen(): boolean;
        /**
         * Subscribe to value changes using the given context.
         * @param fnChange Specifies the event that is invoked when the value is changed.
         * @param pContext Specifies the context.
         * @return Returns a reference to the instance.
         */
        subscribe(
            fnChange: (pValue: ImmutableValue<T, S>) => void,
            pContext?: THookContext
        ): this;
        /**
         * Unsubscribe the given context from the value.
         * @param pContext Specifies the context.
         * @return Returns a reference to the instance.
         */
        unsubscribe(pContext?: THookContext): this;
    }
}

declare module "@tripetto/vfsm/lib/instance/moments/moment" {
    /** Dependencies */
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { IMomentSnapshot } from "@tripetto/vfsm/lib/instance/moments/snapshot";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    export interface IMoment {
        /** Checksum of the moment. */
        readonly checksum: string;
        /** Index of the moment */
        readonly index: number;
        /** Snapshot of the moment. */
        readonly snapshot: IMomentSnapshot;
        /** Reference to the state. */
        readonly state: State;
        /** Reference to the validated transducers. */
        readonly transducers: Transducer[];
        /** Context for the data. */
        readonly context: Context;
        /** Specifies the tense of the moment. */
        readonly tense: "past" | "present" | "future";
        /** Restores the moment. */
        readonly restore: () => void;
    }
    export interface IMoments {
        readonly [id: number]: IMoment | undefined;
    }
    export type TMutableMoment = {
        -readonly [P in keyof IMoment]: IMoment[P];
    };
}

declare module "@tripetto/vfsm/lib/instance/instance" {
    import { IMoment } from "@tripetto/vfsm/lib/instance/moments/moment";
    import { Data } from "@tripetto/vfsm/lib/instance/data/data";
    import { IChainLookup } from "@tripetto/vfsm/lib/chain";
    import { TInstanceEvents } from "@tripetto/vfsm/lib/instance/events";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { IInstanceOrchestrator } from "@tripetto/vfsm/lib/instance/orchestrator";
    import { IInstanceSnapshot } from "@tripetto/vfsm/lib/instance/snapshot";
    import { Item } from "@marblecore/matrix";
    import { Namespace } from "@marblecore/l10n";
    import { Slot } from "@tripetto/slots";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { TThreading } from "@tripetto/vfsm/lib/threading";
    export class Instance extends Item<TInstanceEvents> {
        /**
         * Constructs the instance.
         * @param pOrchestrator Specifies the instance orchestrator.
         * @param sThreading Specifies the threading model (`single` or `multi`).
         * @param pBootstrap Specifies the bootstrap state.
         * @param pSnapshot Specifies an optional snapshot to resume.
         * @param fnInitialize Specifies an optional initializer function that can be used to prefill the instance with data.
         */
        constructor(
            pOrchestrator: IInstanceOrchestrator,
            sThreading: TThreading,
            pBootstrap: State | undefined,
            pSnapshot?: IInstanceSnapshot,
            fnInitialize?: (pInstance: Instance) => void
        );
        /** Retrieves the instance identifier. */
        get id(): string;
        /** Retrieves the namespace identifier. */
        get l10n(): Namespace;
        /** Retrieves the unique ontology fingerprint. */
        get fingerprint(): string;
        /** Retrieves if the instance allows concurrency. */
        get allowConcurrency(): boolean;
        /** Retrieves if the preview mode is enabled. */
        get isPreview(): boolean;
        /** Retrieves if the instance is initialized. */
        get isInitialized(): boolean;
        /** Retrieves if the instance is running. */
        get isRunning(): boolean;
        /** Retrieves if the instance is stopped (interrupted). */
        get isStopped(): boolean;
        /** Retrieves if the instance is being paused. */
        get isPausing(): boolean;
        /** Retrieves if the instance is paused. */
        get isPaused(): boolean;
        /** Retrieves if the instance is at the start. */
        get isAtStart(): boolean;
        /** Retrieves if the instance is at the head. */
        get isAtHead(): boolean;
        /** Retrieves if the instance is at the finish. */
        get isAtFinish(): boolean;
        /** Retrieves if the instance is finishable using the `finish()` method. */
        get isFinishable(): boolean;
        /** Retrieves if the instance is finishing. */
        get isFinishing(): boolean;
        /** Retrieves if the instance has finished. */
        get isFinished(): boolean;
        /** Retrieves the finish reference. */
        get finishReference(): string;
        /** Retrieves if the instance has any data collected. */
        get hasDataCollected(): boolean;
        /** Retrieves the chain lookup object. */
        get lookup(): IChainLookup;
        /** Retrieves the data. */
        get data(): Data;
        /** Retrieves a history of all moments. */
        get history(): IMoment[];
        /** Retrieves the checksum. */
        get checksum(): string;
        /** Retrieves the present moment. */
        get presentMoment(): IMoment | undefined;
        /** Retrieves the instance snapshot. */
        get snapshot(): IInstanceSnapshot;
        /** Retrieves instance error information. */
        get errors(): number;
        /**
         * Opens a transaction that collects all value changes until `closeTransaction` is invoked.
         * @param pTransducer Specifies the transducer that serves as origin.
         * @param fnTransaction Specifies the function that is invoked.
         */
        createTransaction<T = void>(
            pTransducer: Transducer,
            fnTransaction: () => T
        ): void | T;
        /**
         * Reads from the instance cache.
         * @param sKey Specifies the cache key.
         * @return Returns the value;
         */
        readFromCache<T>(sKey: string): T | undefined;
        /**
         * Writes a value to the instance cache.
         * @param sKey Specifies the cache key.
         * @param pValue Specifies the value to cache.
         * @return Returns the value.
         */
        writeToCache<T>(sKey: string, pValue: T): T;
        /**
         * Calculates the stencil hash for the specified type.
         * @param sType Specifies the stencil type (if omitted the exportables stencil is returned).
         * @returns Returns the stencil hash.
         */
        stencil(sType?: "exportables" | "actionables"): string;
        /**
         * Starts the instance.
         * @return Returns `true` if the instance is started.
         */
        start(): boolean;
        /** Restarts the instance. */
        restart(): void;
        /**
         * Stops the instance.
         * @return Returns `true` if the instance was running and is stopped.
         */
        stop(): boolean;
        /**
         * Pauses the instance.
         * @return Returns the instance snapshot necessary to resume the instance or undefined if the instance could not be paused.
         */
        pause(): IInstanceSnapshot | undefined;
        /**
         * Step forward in time.
         * @return Returns `true` if the step succeeded.
         */
        stepForward(): boolean;
        /**
         * Step backward in time.
         * @return Returns `true` if the step succeeded.
         */
        stepBackward(): boolean;
        /**
         * Step to the start.
         * @return Returns `true` if the step succeeded.
         */
        stepToStart(): boolean;
        /**
         * Step to the head of the history.
         * @return Returns `true` if the step succeeded.
         */
        stepToHead(): boolean;
        /**
         * Finish the instance when it is at the end and there are slots collected.
         * @return Returns a promise that resolves when the instance was finished.
         */
        finish(): Promise<void>;
        /**
         * Iterates through all possible transducers and their context.
         * @param fnTransducer Specifies a function that is invoked for each transducer.
         */
        transducers<T extends Transducer, C extends Condition>(
            fnTransducer: (
                pTransducer: T,
                pContext: C[],
                sContext: string
            ) => void
        ): void;
        /**
         * Mark the transducers that are related to the slot as changed.
         * @param pSlot Reference to the slot.
         * @param bOwnerOnly Specifies if only the slot owner needs to be marked (defaults to `false`).
         */
        markAsChanged(pSlot: Slot, bOwnerOnly?: boolean): void;
        /**
         * Retrieves if the specified transducer has changed.
         * @param sId Specifies the transducer identifier.
         * @param bReset Specifies if the change flag needs to be reset.
         */
        hasChanged(sId: string, bReset?: boolean): boolean;
        /**
         * Inherits instance data.
         * @param pInstance Reference to the instance.
         * @returns Returns a reference to the instance.
         */
        inherit(pInstance: Instance): this;
    }
}

declare module "@tripetto/vfsm/lib/interface" {
    /** Dependencies */
    import { ISerialize } from "@marblecore/serializer";
    import { IBranch } from "@tripetto/vfsm/lib/branches/interface";
    export interface IOntologyProperties extends ISerialize {
        /** Contains the property identifier. */
        readonly id: string;
    }
    export interface IOntology extends IBranch {}
}

declare module "@tripetto/vfsm/lib/instance/snapshot" {
    /** Dependencies */
    import { IDataSnapshot } from "@tripetto/vfsm/lib/instance/data/snapshot";
    import { IMomentSnapshot } from "@tripetto/vfsm/lib/instance/moments/snapshot";
    export interface IInstanceSnapshot {
        /** Contains the ontology hash. */
        readonly a: string;
        /** Contains the moment snapshots. */
        readonly b: IMomentSnapshot[];
        /** Contains the data snapshots. */
        readonly c?: IDataSnapshot[];
        /** Contains the index of the active snapshot. */
        readonly d?: number;
        /** Contains if the instance allows concurrency. */
        readonly e?: true;
    }
}

declare module "@tripetto/vfsm/lib/snapshot" {
    /** Dependencies */
    import { IInstanceSnapshot } from "@tripetto/vfsm/lib/instance/snapshot";
    export interface ISnapshot<T = unknown> {
        /** Contains the instance data. */
        readonly a: IInstanceSnapshot[];
        /** Contains custom data to store with the snapshot. */
        readonly b?: T;
    }
}

declare module "@tripetto/vfsm/lib/states/interface" {
    /** Dependencies */
    import { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    export interface IState extends IOntologyProperties {}
}

declare module "@tripetto/vfsm/lib/transducers/interface" {
    /** Dependencies */
    import { ISlot } from "@tripetto/slots";
    import { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    export interface ITransducer extends IOntologyProperties {
        /** Contains the slots for the transducer. */
        readonly slots?: ISlot[];
    }
    export interface ITransducers {
        /** Conditions for the transducers. */
        conditions: Condition[];
        /** Transducers. */
        transducers: (ITransducers | Transducer)[];
    }
}

declare module "@tripetto/vfsm/lib/instance/data/variable" {
    /** Dependencies */
    import { Slot } from "@tripetto/slots";
    import { THookContext } from "@marblecore/hookup";
    import { TSerializeTypes } from "@marblecore/serializer";
    import { IContent } from "@tripetto/vfsm/lib/instance/data/content";
    export interface IVariable {
        /** Contains the identifier of the variable. */
        readonly id: string;
        /** Contains the value as a string. */
        readonly string: string;
        /** Contains the content for the variable. */
        readonly content: IContent[];
        /** Contains if a value is present. */
        readonly hasValue: boolean;
        /** Contains the value. */
        readonly value?: TSerializeTypes;
        /** Contains the value reference. */
        readonly reference?: string;
        /** Reference to the slot. */
        readonly slot?: Slot;
        /** Contains a list of slot references covered by the variable. */
        readonly refs: string[];
        /** Subscribe to variable changes. */
        readonly subscribe: (
            fnChange: (pVariable: IVariable) => void,
            pContext?: THookContext
        ) => void;
        /** Unsubscribe a context from variable changes. */
        readonly unsubscribe: (pContext: THookContext) => void;
    }
}

declare module "@tripetto/vfsm/lib/instance/data/context" {
    import { Slot } from "@tripetto/slots";
    import { IChainLookup } from "@tripetto/vfsm/lib/chain";
    import { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { IVariable } from "@tripetto/vfsm/lib/instance/data/variable";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { Stack } from "@tripetto/vfsm/lib/instance/stack/stack";
    import { Namespace } from "@marblecore/l10n";
    import { TSerializeTypes } from "@marblecore/serializer";
    import { Value } from "@tripetto/vfsm/lib/instance/data/value";
    import { Values } from "@tripetto/vfsm/lib/instance/data/values";
    import { IValueProxy } from "@tripetto/vfsm/lib/instance/data/proxy";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    export interface IContext {
        /** Reference to the condition. */
        readonly condition: Condition;
        /** Specifies the context. */
        readonly context: string;
    }
    export class Context {
        /**
         * Creates a new context.
         * @param pStack Reference to the stack.
         * @param pChain Reference to the chain.
         */
        constructor(pStack: Stack, pChain: IChainLookup);
        /** Retrieves the context instance identifier. */
        get id(): number;
        /** Retrieves the instance of the context. */
        get instance(): Instance;
        /** Retrieves the l10n namespace */
        get l10n(): Namespace;
        /** Retrieves the context key. */
        get key(): string;
        /**
         * Retrieves the validation state of the supplied transducer.
         * @param pTransducer Specifies the transducer.
         * @return Returns the validation state.
         */
        validationOf(pTransducer: Transducer): "unknown" | "pass" | "fail";
        /**
         * Retrieves if this context is the mutable context of the supplied stack.
         * @param pStack Reference to the stack.
         * @return Returns `true` if the context is of the supplied stack.
         */
        isMutableContextOf(pStack: Stack): boolean;
        /**
         * Retrieves the mutable value for the supplied slot instance or identifier.
         * @param pSlotOrId Reference to the slot, value or slot identifier.
         * @return Returns the value or `undefined` if no value was found.
         */
        mutableValueOf<T extends TSerializeTypes, S extends Slot<T> = Slot<T>>(
            pSlotOrId:
                | Slot
                | {
                      slot: Slot;
                  }
                | string
        ): Value<T, S> | undefined;
        /**
         * Retrieves the immutable value for the supplied slot instance or identifier.
         * @param pSlotOrId Reference to the slot, value or slot identifier.
         * @return Returns the value or `undefined` if no value was found.
         */
        immutableValueOf<
            T extends TSerializeTypes,
            S extends Slot<T> = Slot<T>
        >(
            pSlotOrId:
                | Slot
                | {
                      slot: Slot;
                  }
                | string
        ): ImmutableValue<T, S> | undefined;
        /**
         * Retrieves the contextual value for the supplied slot instance or identifier.
         * @param pSlotOrId Reference to the slot, value or slot identifier.
         * @return Returns the value or `undefined` if no value was found.
         */
        contextualValueOf<
            T extends TSerializeTypes,
            S extends Slot<T> = Slot<T>
        >(
            pSlotOrId:
                | Slot
                | {
                      slot: Slot;
                  }
                | string
        ): Value<T, S> | undefined;
        /**
         * Generates a value proxy for the supplied slot instance.
         * @param pValues Reference to the value store.
         * @param pSlot Reference to the slot.
         * @param bPipe Specifies if the variable if part of a pipe.
         * @param sSeparator Specifies the value separator.
         * @param fnToString Optional function that is used to cast the variable to
         * a string.
         * @return Returns a proxy object.
         */
        proxyValueOf(
            pValues: Values,
            pSlot: Slot,
            bPipe: boolean,
            sSeparator: string,
            fnToString?: (
                pValue: ImmutableValue,
                pContext: Context
            ) => string | undefined
        ): IValueProxy;
        /**
         * Retrieves the variable for the supplied slot instance or (pipe) identifier.
         * @param pSlotOrId Reference to the slot, value or slot identifier.
         * @param fnPipe Specifies the function used to retrieve the pipe identifier.
         * @param fnToString Optional function that is used to cast the variable to
         * a string.
         * @param finishReferenceIdentifier Specifies the identifier that retrieves
         * the finish reference.
         * @param sSeparator Specifies the separator used for multiple values in a pipe.
         * @param bPipe Specifies the variable is part of a pipe.
         * @return Returns the variable or `undefined` if the variable was not found,
         * not valid or not set.
         */
        variableFor(
            pSlotOrId:
                | Slot
                | {
                      slot: Slot;
                  }
                | string,
            fnPipe?: (
                pCondition: Condition,
                pContext: Context
            ) => string | undefined,
            fnToString?: (
                pValue: ImmutableValue,
                pContext: Context
            ) => string | undefined,
            finishReferenceIdentifier?: string,
            sSeparator?: string,
            bPipe?: boolean
        ): IVariable | undefined;
    }
}

declare module "@tripetto/vfsm/lib/instance/data/data" {
    import { Matrix } from "@marblecore/matrix";
    import { Slot } from "@tripetto/slots";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { IInstanceOrchestrator } from "@tripetto/vfsm/lib/instance/orchestrator";
    import { TSerializeTypes } from "@marblecore/serializer";
    import {
        IDataChangeEvent,
        TValuesEvents,
    } from "@tripetto/vfsm/lib/instance/data/events";
    import { Values } from "@tripetto/vfsm/lib/instance/data/values";
    export class Data extends Matrix<Values, TValuesEvents> {
        /**
         * Creates a new data instance.
         * @param pInstance Reference to the instance.
         * @param pOrchestrator Specifies the instance orchestrator.
         * @param fnUpdate Specifies the update handler.
         */
        constructor(
            pInstance: Instance,
            pOrchestrator: IInstanceOrchestrator,
            fnUpdate: (pEvent: IDataChangeEvent) => void
        );
        /** Retrieves all exportable values. */
        get exportables(): Values[];
        /** Retrieves if there are sealed values collected. */
        get hasSealedValues(): boolean;
        /**
         * Retrieve the values for the supplied slot.
         * @param pSlotOrId Reference to the slot, value or slot identifier.
         * @return Returns the values or `undefined` if no values are found.
         */
        forSlot<T extends TSerializeTypes, S extends Slot<T>>(
            pSlotOrId:
                | Slot
                | {
                      slot: Slot;
                  }
                | string
        ): Values<T, S> | undefined;
        /** Freeze the data. */
        freeze(): void;
        /** Unfreeze the data. */
        unfreeze(): void;
    }
}

declare module "@tripetto/vfsm/lib/ontology" {
    import { Namespace } from "@marblecore/l10n";
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { Errors } from "@tripetto/vfsm/enum/errors";
    import { IConstructors } from "@tripetto/vfsm/lib/constructors";
    import { IInstanceSnapshot } from "@tripetto/vfsm/lib/instance/snapshot";
    import { Matrix } from "@marblecore/matrix";
    import { IMetrics } from "@tripetto/vfsm/lib/metrics";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { TOntologyEvents } from "@tripetto/vfsm/lib/events";
    import { ISnapshot } from "@tripetto/vfsm/lib/snapshot";
    import { Main } from "@tripetto/vfsm/lib/states/main";
    import { Slot } from "@tripetto/slots";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { States } from "@tripetto/vfsm/lib/states/states";
    import { TBranch } from "@tripetto/vfsm/lib/branches/type";
    import { TCondition } from "@tripetto/vfsm/lib/conditions/type";
    import { THooks } from "@marblecore/hookup";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { TState } from "@tripetto/vfsm/lib/states/type";
    import { TThreading } from "@tripetto/vfsm/lib/threading";
    import { TTransducer } from "@tripetto/vfsm/lib/transducers/type";
    import { IDataChangeEvent } from "@tripetto/vfsm/lib/instance/data/events";
    export type TOntology = Ontology<State, Transducer, Branch, Condition>;
    export abstract class Ontology<
        S extends State,
        T extends Transducer,
        B extends Branch,
        C extends Condition,
        O extends {} = {},
        H extends THooks = void
    > extends Matrix<Instance, TOntologyEvents | H> {
        /**
         * Constructs a new ontology.
         * @param pState Reference to the state constructor.
         * @param pTransducer Reference to the transducer constructor.
         * @param pBranch Reference to the branch constructor.
         * @param pCondition Reference to the condition constructor.
         * @param pOntology Reference to the ontology data.
         * @param fnFingerprint Specifies a fingerprint generator.
         * @param fnStencil Specifies a stencil validator.
         * @param bAutoPushForward Specifies if auto push-forward should be used. when
         * enabled, steps forward are made automatically as long as there are no
         * validation obstructions (disabled by default).
         * @param bCalculateFingerprint Specifies if the fingerprint should be
         * calculated (enabled by default). Disabling this calculation can improve
         * performance in preview/test mode.
         * @param sPreview Specifies if the preview mode is enabled. In this mode all
         * conditions and validations will pass without evaluating them. This allows
         * an application to iterate through all possible states and transducers. Can
         * be one of the following values:
         * - `disabled`: The preview mode is disabled;
         * - `compact`: Only show unique branches in the preview;
         * - `enabled`: Show every possible branch.
         * @param pL10 Reference to a l10n namespace to use.
         * @param bVerbose Enables or disables verbose logging (disabled by default).
         * @return Returns a reference to the instance.
         */
        constructor(
            pState: TState<S>,
            pTransducer: TTransducer<T>,
            pBranch: TBranch<B>,
            pCondition: TCondition<C>,
            pOntology?: string | O,
            fnFingerprint?: (pReference: S | T | B | C | Slot) => string,
            fnStencil?: (pReference: T) => boolean,
            bAutoPushForward?: boolean,
            bCalculateFingerprint?: boolean,
            sPreview?: "disabled" | "compact" | "enabled",
            pL10?: Namespace,
            bVerbose?: boolean
        );
        /** Retrieves the constructors. */
        get constructors(): IConstructors<S, T, B, C>;
        /** Retrieves the ontology data. */
        get ontology(): Readonly<O> | undefined;
        /** Retrieves the unique ontology fingerprint. */
        get fingerprint(): string;
        /** Reference to the l10n namespace. */
        get l10n(): Namespace;
        /** Retrieves a reference to the state dispatcher. */
        get states(): States;
        /** Retrieves a reference to the main state. */
        get main(): Main | undefined;
        /** Retrieves if the ontology is loaded successfully. */
        get isLoaded(): boolean;
        /** Retrieves if the VFSM is running. */
        get isRunning(): boolean;
        /** Retrieves if the VFSM is stopped. */
        get isStopped(): boolean;
        /** Retrieves if the VFSM is about to finish */
        get isFinishing(): boolean;
        /** Retrieves if the VFSM has finished (without pausing or stopping it). */
        get isFinished(): boolean;
        /** Retrieves if the VFSM is pausing. */
        get isPausing(): boolean;
        /** Retrieves if the VFSM is paused. */
        get isPaused(): boolean;
        /** Retrieves if the preview mode is enabled. */
        get isPreview(): boolean;
        /** Retrieves if verbose logging is enabled. */
        get isVerbose(): boolean;
        /** Enables or disables verbose logging. */
        set isVerbose(bVerbose: boolean);
        /** Retrieves if automatic push-forward is enabled. */
        get autoPushForward(): boolean;
        /** Enables or disables automatic push-forward. */
        set autoPushForward(bAutoPushForward: boolean);
        /** Retrieves the ontology metrics. */
        get metrics(): Readonly<IMetrics>;
        /** Retrieves the log message list. */
        get log(): string[];
        /** Retrieves the logger function. */
        get logger(): (sMessage: string) => void;
        /** Sets the log function. */
        set logger(fnLogger: (sMessage: string) => void);
        /**
         * Invoked when an instance is created.
         * @param pInstance Reference to the instance.
         * @event
         */
        protected onInstanceCreate(pInstance: Instance): void;
        /**
         * Invoked when an instance is started.
         * @param pInstance Reference to the instance.
         * @event
         */
        protected onInstanceStart(pInstance: Instance): void;
        /**
         * Invoked when the collected data of an instance is changed.
         * @param pInstance Reference to the instance.
         * @param pDataChangeEvent Reference to a data change event.
         * @event
         */
        protected onInstanceData(
            pInstance: Instance,
            pDataChangeEvent: IDataChangeEvent
        ): void;
        /**
         * Invoked when an instance is processing.
         * @param pInstance Reference to the instance.
         * @param bProcessing Specifies if the instance is processing.
         * @event
         */
        protected onInstanceProcessing(
            pInstance: Instance,
            bProcessing: boolean
        ): void;
        /**
         * Invoked when an instance wants to check if it can step forward.
         * @param pInstance Reference to the instance.
         * @param pState Reference to the current state.
         * @return Returns `true` if step forward is allowed.
         * @event
         */
        protected onInstanceStepForward(
            pInstance: Instance,
            pState: State
        ): boolean;
        /**
         * Invoked when the instance is updated.
         * @param pInstance Reference to the instance.
         * @event
         */
        protected onInstanceUpdate(pInstance: Instance): void;
        /**
         * Invoked when an instance ends. This event is always invoked
         * when an instance is finished, stopped or paused.
         * @param pInstance Reference to the instance.
         * @param sType Specifies why the instance has ended. Can be one of the
         * following values:
         * - `finished`: The instance was finished;
         * - `stopped`: The instance was forced to stop;
         * - `paused`: The instance was paused.
         * @event
         */
        protected onInstanceEnd(
            pInstance: Instance,
            sType: "finished" | "stopped" | "paused"
        ): void;
        /**
         * Invoked when an instance is about to finish.
         * @param pInstance Reference to the instance.
         * @param fnFinish A callback function that can be used to control the finish process when you override this event and return `false`.
         * @return Return `true` if the instance finish is allowed or `false` if the instance finish should wait till the callback function is invoked.
         * @event
         */
        protected onInstanceFinishing(
            pInstance: Instance,
            fnFinish: (bFinish: boolean, sReference?: string) => void
        ): boolean;
        /**
         * Invoked when an instance has finished.
         * @param pInstance Reference to the instance.
         * @event
         */
        protected onInstanceFinish(pInstance: Instance): void;
        /**
         * Invoked when an instance is paused.
         * @param pInstance Reference to the instance.
         * @event
         */
        protected onInstancePause(pInstance: Instance): void;
        /**
         * Invoked when an instance is stopped.
         * @param pInstance Reference to the instance.
         * @event
         */
        protected onInstanceStop(pInstance: Instance): void;
        /**
         * Spawns a new instance. Throws an error if no ontology is loaded.
         * @param sThreading Specifies if a single or multi threaded instance should be spawned.
         * @param pSnapshotOrCallback Specifies the optional instance snapshot.
         * @param fnInitialize Specifies an optional initializer function that can be used to prefill the instance with data.
         * @return Returns a reference to the instance.
         */
        protected spawn(
            sThreading?: TThreading,
            pSnapshot?: IInstanceSnapshot,
            fnInitialize?: (pInstance: Instance) => void
        ): Instance;
        /**
         * Calculates the stencil hash for the specified type.
         * @param sType Specifies the stencil type (if omitted the exportables stencil is returned).
         * @returns Returns the stencil hash.
         */
        stencil(sType?: "exportables" | "actionables"): string;
        /**
         * Loads an ontology.
         * @param pOntology Specifies the ontology data.
         * @return Returns `Errors.None` if the ontology is loaded.
         */
        load(pOntology: O | string): Errors;
        /** Unloads the ontology. */
        unload(): void;
        /**
         * Starts a new instance. Throws an error if no ontology is loaded.
         * @param pSnapshot Specifies the optional instance snapshot.
         * @param fnInitialize Specifies an optional initializer function that can be used to prefill the instance with data.
         * @return Returns a reference to the instance.
         */
        start(
            pSnapshot?: IInstanceSnapshot,
            fnInitialize?: (pInstance: Instance) => void
        ): Instance;
        /**
         * Stops all running instances.
         * @return Returns `true` if there was an actual instance running.
         */
        stop(): boolean;
        /**
         * Takes a snapshot of the VSFM ontology.
         * @param pCustomData Specifies custom data to store in the snapshot.
         * @return Returns the snapshot which is necessary to resume the VFSM
         * or `undefined` if no instance was active.
         */
        snapshot<Data>(pCustomData?: Data): ISnapshot<Data> | undefined;
        /**
         * Pauses the VFSM ontology.
         * @param pCustomData Specifies custom data to store in the snapshot.
         * @return Returns the snapshot which is necessary to resume the VFSM
         * or `undefined` if no instance was running.
         */
        pause<Data>(pCustomData?: Data): ISnapshot<Data> | undefined;
        /**
         * Pauses the VFSM ontology.
         * @param pCustomData Specifies custom data to store in the snapshot.
         * @param fnPausing Specifies a function that is invoked when the pause is
         * initiated. This function receives the pause snapshot together with a
         * callback function that needs to be invoked to indicate if the pause
         * should be executed or dismissed.
         * @return Returns a promise with the pause snapshot.
         */
        pause<Data>(
            pCustomData: Data | undefined,
            fnPausing: (
                pSnapshot: ISnapshot<Data>,
                fnDone: (bPause: boolean) => void
            ) => void
        ): Promise<ISnapshot<Data>>;
        /**
         * Resumes the VFSM ontology. This will resume all instances that are
         * present in the snapshot data.
         * @param pSnapshot Specifies the snapshot data.
         * @return Returns `true` if the resume is successful and the VFSM is
         * running.
         */
        resume<Data>(pSnapshot: ISnapshot<Data>): boolean;
        /**
         * Enables or disables the preview mode. Any running instances are updated.
         * @param sMode Specifies the preview mode. Can be one of the following values:
         * - `disabled`: The preview mode is disabled;
         * - `compact`: Only show unique branches in the preview;
         * - `enabled`: Show every possible branch.
         * @param bRestart Specifies if the instances should be restarted.
         */
        previewMode(
            sMode?: "disabled" | "compact" | "enabled",
            bRestart?: boolean
        ): "disabled" | "compact" | "enabled";
        /**
         * Iterates through each item in the ontology.
         * @param fnItem Callback which is invoked for each item.
         * @param sType Specifies the type of items.
         * @return Returns `true` if the iteration was interrupted.
         */
        forEach<Types extends State | Transducer | Branch | Condition>(
            fnItem: (pItem: Types) => boolean | void,
            sType?:
                | "states"
                | "transducers"
                | "branches"
                | "conditions"
                | "all",
            pExclude?: State | Transducer | Branch | Condition
        ): boolean;
        /** Destroys the ontology. */
        destroy(): void;
    }
    export class OntologyProcessor<
        S extends State,
        T extends Transducer,
        B extends Branch,
        C extends Condition,
        O extends {} = {}
    > extends Ontology<S, T, B, C, O> {
        constructor(
            pState: TState<S>,
            pTransducer: TTransducer<T>,
            pBranch: TBranch<B>,
            pCondition: TCondition<C>,
            pOntology?: string | O,
            fnFingerprint?: (pReference: S | T | B | C | Slot) => string,
            fnStencil?: (pReference: T) => boolean
        );
    }
}

declare module "@tripetto/vfsm/lib/instance/data/value" {
    import { THookContext } from "@marblecore/hookup";
    import { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { Slot } from "@tripetto/slots";
    import { TSerializeTypes } from "@marblecore/serializer";
    export class Value<
        T extends TSerializeTypes = TSerializeTypes,
        S extends Slot<T> = Slot<T>
    > {
        /**
         * Creates a new value store.
         * @param pSlot Reference to the slot.
         * @param sContext Contains the context of the value.
         * @param pContext Contains a reference to the context.
         * @param fnChange Contains the change callback.
         * @param pValue Contains the hot value.
         * @param sReference Contains the value reference.
         * @param nTime Contains the UTC set time of the value.
         * @param bSealed Specifies if the value was sealed.
         * @param bLocked Specifies if the value was locked.
         */
        constructor(
            pSlot: S,
            sContext: string,
            pContext: Context | undefined,
            fnChange: (pValue: Value<T, S>, sChange: "value" | "seal") => void,
            pValue?: TSerializeTypes,
            sReference?: string,
            nTime?: number,
            bSealed?: boolean,
            bLocked?: boolean
        );
        /** Retrieves the value instance identifier. */
        get id(): number;
        /** Retrieves a unique key for the value within its context (the key is prefixed with an underscore). */
        get key(): string;
        /** Retrieves the context for the value. */
        get context(): "*" | string;
        /** Retrieves a reference to the context. */
        get contextRef(): Context | undefined;
        /** Retrieves a reference to the slot. */
        get slot(): S;
        /** Retrieves the pristine value. */
        get pristine(): TSerializeTypes;
        /** Sets the pristine value. */
        set pristine(pValue: TSerializeTypes);
        /** Retrieves the value. */
        get value(): T;
        /** Sets the value. */
        set value(pValue: T);
        /**
         * Retrieves the value as a string (this will always return a string with a
         * value regardless of a pristine value was set).
         */
        get string(): string;
        /** Retrieves the optional value reference. */
        get reference(): string | undefined;
        /** Retrieves if there is a value. */
        get hasValue(): boolean;
        /** Retrieves if the value is modified. */
        get isModified(): boolean;
        /** Retrieves if the value is sealed. */
        get isSealed(): boolean;
        /** Specifies if the value is sealed. */
        set isSealed(bSealed: boolean);
        /** Retrieves if the value is dirty. */
        get isDirty(): boolean;
        /** Retrieves if the seal should to be verified. */
        get shouldVerifySeal(): boolean;
        /** Retrieves if the value is frozen. */
        get isFrozen(): boolean;
        /** Retrieves if the value is locked. */
        get isLocked(): boolean;
        /** Retrieves if the value is awaiting. */
        get isAwaiting(): boolean;
        /** Specifies if the value can be prefilled. */
        get allowPrefill(): boolean;
        /** Retrieves the value set timestamp. */
        get time(): number | undefined;
        /** Retrieves a immutable representation of the value. */
        get immutable(): ImmutableValue<T, S>;
        /** Precursor a value. */
        precursor(): void;
        /**
         * Sets the pristine value and optional reference.
         * @param pValue Specifies the pristine value.
         * @param sReference Specifies an optional value reference.
         * @return Returns the value.
         */
        set(pValue: TSerializeTypes, sReference?: string): T;
        /**
         * Concatenates a value.
         * @param sValue Specifies the value to concatenate.
         * @param pContext Specifies the context of the value.
         * @param sSeparator Specifies a separator to use.
         * @returns Returns the concatenated string.
         */
        concatenate(
            sValue: string,
            pContext?: {} | string,
            sSeparator?: string
        ): string;
        /**
         * Sets the output of an operation as value.
         * @param pValue Specifies the result of the operation.
         * @param pContext Specifies the context of the operation.
         * @returns Returns a reference to the value.
         */
        operation(pValue: TSerializeTypes, pContext: {} | string): this;
        /**
         * Resets the value and optional reference to the prefill/initial value.
         * @param pPrefillValue Specifies the prefill value.
         * @param sPrefillReference Specifies an optional prefill value reference.
         * @return Retrieves the value.
         */
        reset(pPrefillValue?: TSerializeTypes, sPrefillReference?: string): T;
        /**
         * Sets a default value and optional reference.
         * @param pValue Specifies the pristine value.
         * @param sReference Specifies an optional value reference.
         * @return Retrieves the value.
         */
        default(pValue: TSerializeTypes, sReference?: string): T;
        /** Awaits a value. */
        await(): this;
        /** Cancels an await operation. */
        cancelAwait(): this;
        /** Freezes a value. */
        freeze(): void;
        /** Unfreezes a value. */
        unfreeze(): void;
        /** Locks a value. */
        lock(): void;
        /** Unlocks a value. */
        unlock(): void;
        /** Clears a value. */
        clear(): T;
        /**
         * Imports a value.
         * @param pValue Specifies the pristine value.
         * @param sReference Specifies an optional value reference.
         * @param nTime Specifies the set timestamp.
         * @param bSilent Specifies if the import is silent and should not emit event.
         * @return Retrieves the value.
         */
        import(
            pValue: TSerializeTypes,
            sReference?: string,
            nTime?: number,
            bSilent?: boolean
        ): T;
        /**
         * Marks the value as confirmed. In this case it is not necessary to
         * explicit set a value to make the value valid.
         * @return Returns a reference to the instance.
         */
        confirm(): this;
        /**
         * Marks the value as dirty. An explicit value should be set to mark
         * the value as valid or `confirm()` should be invoked.
         * @return Returns a reference to the instance.
         */
        dirty(): this;
        /**
         * Subscribe to value changes using the given context.
         * @param fnChange Specifies the event that is invoked when the value is changed.
         * @param pContext Specifies the context.
         * @return Returns a reference to the instance.
         */
        subscribe(
            fnChange: (pValue: Value<T, S>) => void,
            pContext?: THookContext,
            sMode?: "asynchronous" | "synchronous"
        ): this;
        /**
         * Unsubscribe the given context from the value.
         * @param pContext Specifies the context.
         * @return Returns a reference to the instance.
         */
        unsubscribe(pContext?: THookContext): this;
        /**
         * Listens to value changes.
         * @param fnListener Specifies the event that is invoked when the value is changed.
         * @return Returns a reference to the instance.
         */
        useListener(fnListener: (pValue: Value<T, S>) => void): this;
        /**
         * Specifies a modifier function to use.
         * @param fnModifier Specifies the modifier function.
         * @return Returns a reference to the instance.
         */
        useModifier(
            fnModifier:
                | ((pValue: Value<T, S>) =>
                      | {
                            readonly value: TSerializeTypes;
                            readonly reference?: string;
                        }
                      | undefined)
                | undefined
        ): this;
        /**
         * Attaches a context to the value.
         * @param pContext Reference to the context
         * @returns Returns a reference to the instance.
         */
        attachContext(pContext: Context | undefined): this;
        /**
         * Specifies a context function to use.
         * @param fnContext Specifies the context function.
         * @return Returns a reference to the instance.
         */
        useContext(
            fnContext:
                | ((pValue: Value<T, S>, pContext: Context) => void)
                | undefined
        ): this;
        /**
         * Mounts a new context.
         * @param pContext Reference to the context.
         * @return Returns a reference to the instance.
         */
        mountContext(pContext: Context): this;
    }
}

declare module "@tripetto/vfsm/lib/instance/data/values" {
    /** Dependencies */
    import { Item } from "@marblecore/matrix";
    import { TSerializeTypes } from "@marblecore/serializer";
    import {
        IDataChangeEvent,
        TValuesEvents,
    } from "@tripetto/vfsm/lib/instance/data/events";
    import { Slot } from "@tripetto/slots";
    import { Data } from "@tripetto/vfsm/lib/instance/data/data";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { IInstanceOrchestrator } from "@tripetto/vfsm/lib/instance/orchestrator";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { IContent } from "@tripetto/vfsm/lib/instance/data/content";
    import {
        IDataImport,
        IDataValue,
        IDataValues,
    } from "@tripetto/vfsm/lib/instance/data/interface";
    import { IDataSnapshot } from "@tripetto/vfsm/lib/instance/data/snapshot";
    import { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    import { Stack } from "@tripetto/vfsm/lib/instance/stack/stack";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { Value } from "@tripetto/vfsm/lib/instance/data/value";
    export class Values<
        T extends TSerializeTypes = TSerializeTypes,
        S extends Slot<T> = Slot<T>
    > extends Item<TValuesEvents> {
        /**
         * Constructs a new values store.
         * @param pTransducer Reference to the transducer.
         * @param pInstance Reference to the instance.
         * @param pOrchestrator Specifies the instance orchestrator.
         * @param pSlot Reference to the slot.
         */
        constructor(
            pTransducer: Transducer,
            pSlot: S,
            pInstance: Instance,
            pOrchestrator: IInstanceOrchestrator,
            fnUpdate: (pEvent: IDataChangeEvent) => void
        );
        /** Retrieves a reference to the data instance. */
        get data(): Data;
        /** Retrieves a reference to the transducer. */
        get transducer(): Transducer;
        /** Retrieves a reference to the slot. */
        get slot(): S;
        /** Retrieves the values. */
        get values(): IDataValues;
        /** Retrieves if there are sealed values collected. */
        get hasSealedValues(): boolean;
        /**
         * Retrieves the most relevant value for the supplied stack.
         * @param pStack Reference to the stack to use as a context.
         * @param bContextual Specifies if the value is contextual (local).
         * @param pContext Reference to the context instance.
         * @return Returns a mutable value.
         */
        mutableValue(
            pStack: Stack | undefined,
            bContextual: boolean,
            pContext: Context
        ): Value<T, S>;
        /**
         * Retrieves the most relevant value for the supplied stack.
         * @param pStack Reference to the stack to use as a context.
         * @return Returns an immutable value.
         */
        immutableValue(
            pStack: Stack | undefined
        ): ImmutableValue<T, S> | undefined;
        /**
         * Retrieves the most relevant values for the supplied stack.
         * @param pStack Reference to the stack to use as a context.
         * @return Returns an array of immutable values.
         */
        immutableValues(pStack: Stack | undefined): ImmutableValue<T, S>[];
        /**
         * Retrieves if the value of the specified context was modified.
         * @param sContext Specifies the context.
         * @return Returns `true` if the value was modified.
         */
        isModified(sContext: string): boolean;
        /**
         * Snapshots the store.
         * @return Returns the store snapshot.
         */
        snapshot(): IDataSnapshot | undefined;
        /**
         * Restores the values.
         * @param pSnapshot Specifies the snapshot to restore.
         */
        restore(pSnapshot: IDataSnapshot): void;
        /** Freeze the data. */
        freeze(): void;
        /** Unfreeze the data. */
        unfreeze(): void;
        /**
         * Imports a single value.
         * @param sContext Specifies the context.
         * @param pData Specifies the data to import.
         */
        importValue(sContext: string, pData: Partial<IDataValue>): void;
        /**
         * Import a values object.
         * @param pValues Specifies the values to import.
         */
        import(pValues: IDataImport): void;
        /** Retrieves if there is any value set. */
        hasValue(pStack: Stack | undefined): boolean;
        /** Retrieves the value of the value set. */
        getValue(pStack: Stack | undefined): number | boolean | undefined;
        /** Retrieves all string values. */
        getString(
            pStack: Stack | undefined,
            pContext: Context,
            bPipe: boolean,
            fnToString?: (
                pValue: ImmutableValue<T, S>,
                pContext: Context
            ) => string | undefined
        ): string[];
        /** Retrieves the concatenated content for all values. */
        getContent(
            pStack: Stack | undefined,
            pContext: Context,
            bPipe: boolean,
            sSeparator: string,
            fnToString?: (
                pValue: ImmutableValue<T, S>,
                pContext: Context
            ) => string | undefined
        ): IContent[];
        /** Retrieves a list of values refs */
        getRefs(pStack: Stack | undefined, bPipe: boolean): string[];
        /**
         * Inherit instance data.
         * @param pInheritor Specifies the instance that inherits the data.
         */
        inherit(pInheritor: Instance): void;
    }
}

declare module "@tripetto/vfsm/enum/errors" {
    export enum Errors {
        /** No errors were detected. */
        None = 0,
        /** The state data in the ontology is invalid. */
        OntologyInvalidState = 1,
        /** The transducer data in the ontology is invalid. */
        OntologyInvalidTransducer = 2,
        /** The branch data in the ontology is invalid. */
        OntologyInvalidBranch = 4,
        /** The condition data in the ontology is invalid. */
        OntologyInvalidCondition = 8,
        /** The jump reference in the ontology is invalid. */
        OntologyInvalidJump = 16,
        /** The resume data is invalid. */
        ResumeInvalidHash = 32,
        /** One of more resume data slots are invalid. */
        ResumeInvalidData = 64,
        /** The resume stack is invalid. */
        ResumeInvalidStack = 128,
    }
}

declare module "@tripetto/vfsm/lib/decorator" {
    import { TBranch } from "@tripetto/vfsm/lib/branches/type";
    import { TCondition } from "@tripetto/vfsm/lib/conditions/type";
    import { TState } from "@tripetto/vfsm/lib/states/type";
    import { TTransducer } from "@tripetto/vfsm/lib/transducers/type";
    export type TTypes = TState | TTransducer | TBranch | TCondition;
    /**
     * Sets the property name for an ontology.
     * @param sProperty Specifies the property name.
     */
    export function ontology<T extends TTypes>(
        sProperty: string
    ): (pType: T) => void;
}

declare module "@tripetto/vfsm/lib/events" {
    /** Dependencies */
    import { Data } from "@tripetto/vfsm/lib/instance/data/data";
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { Slot } from "@tripetto/slots";
    import { TOntology } from "@tripetto/vfsm/lib/ontology";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    export type TOntologyEvents =
        | "OnLoad"
        | "OnUnload"
        | "OnStart"
        | "OnEnd"
        | "OnPause"
        | "OnResume"
        | "OnStop"
        | "OnFinish"
        | "OnInstanceCreate"
        | "OnInstanceStart"
        | "OnInstanceData"
        | "OnInstanceProcessing"
        | "OnInstanceEnd"
        | "OnInstancePause"
        | "OnInstanceStop"
        | "OnInstanceFinish";
    export interface IOntologyEvent<T extends THooks> extends IHookPayload<T> {
        readonly ontology: TOntology;
    }
    export interface IOntologyLoadEvent extends IOntologyEvent<"OnLoad"> {}
    export interface IOntologyUnloadEvent extends IOntologyEvent<"OnUnload"> {}
    export interface IOntologyStartEvent extends IOntologyEvent<"OnStart"> {}
    export interface IOntologyPauseEvent extends IOntologyEvent<"OnPause"> {}
    export interface IOntologyStopEvent extends IOntologyEvent<"OnStop"> {}
    export interface IOntologyFinishEvent extends IOntologyEvent<"OnFinish"> {}
    export interface IOntologyResumeEvent<T>
        extends IOntologyEvent<"OnResume"> {
        readonly data?: T;
    }
    export interface IOntologyEndEvent extends IOntologyEvent<"OnEnd"> {
        readonly type: "finished" | "stopped" | "paused";
    }
    export interface IOntologyInstanceEvent<T extends THooks>
        extends IOntologyEvent<T> {
        readonly instance: Instance;
    }
    export interface IOntologyInstanceCreateEvent
        extends IOntologyInstanceEvent<"OnInstanceCreate"> {}
    export interface IOntologyInstanceStartEvent
        extends IOntologyInstanceEvent<"OnInstanceStart"> {}
    export interface IOntologyInstancePauseEvent
        extends IOntologyInstanceEvent<"OnInstancePause"> {}
    export interface IOntologyInstanceStopEvent
        extends IOntologyInstanceEvent<"OnInstanceStop"> {}
    export interface IOntologyInstanceFinishEvent
        extends IOntologyInstanceEvent<"OnInstanceFinish"> {}
    export interface IOntologyInstanceDataEvent
        extends IOntologyInstanceEvent<"OnInstanceData"> {
        readonly transducer: Transducer;
        readonly slot: Slot;
        readonly data: Data;
        readonly id: number;
        readonly context: string;
        readonly value: ImmutableValue;
    }
    export interface IOntologyInstanceProcessingEvent
        extends IOntologyInstanceEvent<"OnInstanceProcessing"> {
        readonly processing: boolean;
    }
    export interface IOntologyInstanceEndEvent
        extends IOntologyInstanceEvent<"OnInstanceEnd"> {
        readonly type: "finished" | "stopped" | "paused";
    }
}

declare module "@tripetto/vfsm/lib/instance/events" {
    /** Dependencies */
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    export type TInstanceEvents =
        | "OnCreate"
        | "OnStart"
        | "OnEnd"
        | "OnPause"
        | "OnStop"
        | "OnFinish";
    export interface IInstanceEvent<T extends THooks> extends IHookPayload<T> {
        readonly instance: Instance;
    }
    export interface IInstanceCreateEvent extends IInstanceEvent<"OnCreate"> {}
    export interface IInstanceStartEvent extends IInstanceEvent<"OnStart"> {}
    export interface IInstancePauseEvent extends IInstanceEvent<"OnPause"> {}
    export interface IInstanceStopEvent extends IInstanceEvent<"OnStop"> {}
    export interface IInstanceFinishEvent extends IInstanceEvent<"OnFinish"> {}
    export interface IInstanceEndEvent extends IInstanceEvent<"OnEnd"> {
        readonly type: "finished" | "stopped" | "paused";
    }
}

declare module "@tripetto/vfsm/lib/branches/events" {
    /** Dependencies */
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    export type TBranchEvents =
        | "OnReady"
        | "OnEnter"
        | "OnSkip"
        | "OnLeave"
        | "OnCancel";
    export interface IBranchEvent<T extends THooks> extends IHookPayload<T> {
        readonly branch: Branch;
        readonly context: Context;
    }
    export interface IBranchReadyEvent extends IBranchEvent<"OnReady"> {}
    export interface IBranchEnterEvent extends IBranchEvent<"OnEnter"> {}
    export interface IBranchSkipEvent extends IBranchEvent<"OnSkip"> {}
    export interface IBranchLeaveEvent extends IBranchEvent<"OnLeave"> {}
    export interface IBranchCancelEvent extends IBranchEvent<"OnCancel"> {}
}

declare module "@tripetto/vfsm/lib/states/events" {
    /** Dependencies */
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { State } from "@tripetto/vfsm/lib/states/state";
    export type TStateEvents =
        | "OnEnter"
        | "OnSkip"
        | "OnLeave"
        | "OnCancel"
        | "OnValidated"
        | "OnEvaluate";
    export interface IStateEvent<T extends THooks> extends IHookPayload<T> {
        readonly state: State;
        readonly context: Context;
    }
    export interface IStateEnterEvent extends IStateEvent<"OnEnter"> {}
    export interface IStateSkipEvent extends IStateEvent<"OnSkip"> {}
    export interface IStateLeaveEvent extends IStateEvent<"OnLeave"> {}
    export interface IStateCancelEvent extends IStateEvent<"OnCancel"> {}
    export interface IStateEvaluateEvent extends IStateEvent<"OnEvaluate"> {
        readonly evaluating: boolean;
    }
    export interface IStateValidatedEvent extends IStateEvent<"OnValidated"> {
        readonly result: "fail" | "pass";
        readonly type: "initial" | "revalidate";
    }
}

declare module "@tripetto/vfsm/lib/transducers/events" {
    /** Dependencies */
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    export type TTransducerEvents =
        | "OnEnter"
        | "OnSkip"
        | "OnLeave"
        | "OnCancel"
        | "OnValidated";
    export interface ITransducerEvent<T extends THooks>
        extends IHookPayload<T> {
        readonly transducer: Transducer;
        readonly context: Context;
    }
    export interface ITransducerEnterEvent
        extends ITransducerEvent<"OnEnter"> {}
    export interface ITransducerSkipEvent extends ITransducerEvent<"OnSkip"> {}
    export interface ITransducerLeaveEvent
        extends ITransducerEvent<"OnLeave"> {}
    export interface ITransducerCancelEvent
        extends ITransducerEvent<"OnCancel"> {}
    export interface ITransducerValidatedEvent
        extends ITransducerEvent<"OnValidated"> {
        readonly result: "fail" | "pass";
        readonly type: "initial" | "revalidate";
    }
}

declare module "@tripetto/vfsm/lib/instance/data/events" {
    /** Dependencies */
    import { Data } from "@tripetto/vfsm/lib/instance/data/data";
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { Slot } from "@tripetto/slots";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { TSerializeTypes } from "@marblecore/serializer";
    import { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    export type TValuesEvents = "OnValueChange" | "OnSealChange";
    interface IChangeEvent<
        T extends TSerializeTypes,
        S extends Slot<T>,
        H extends THooks
    > extends IHookPayload<H> {
        /** Reference to the transducer to whom the value belongs. */
        readonly transducer: Transducer;
        /** Reference to the slot. */
        readonly slot: S;
        /** Reference to the data instance that holds the all values. */
        readonly data: Data;
        /** Identifier of the value. */
        readonly id: number;
        /** Contains the key of the context. */
        readonly context: string;
        /** Reference to the value. */
        readonly value: ImmutableValue<T, S>;
    }
    export interface IValueChangeEvent<
        T extends TSerializeTypes = TSerializeTypes,
        S extends Slot<T> = Slot<T>
    > extends IChangeEvent<T, S, "OnValueChange"> {}
    export interface ISealChangeEvent<
        T extends TSerializeTypes = TSerializeTypes,
        S extends Slot<T> = Slot<T>
    > extends IChangeEvent<T, S, "OnSealChange"> {}
    export type IDataChangeEvent = Omit<
        IValueChangeEvent,
        "hook" | "context"
    > & {
        /** Contains the key of the context. */
        readonly key: string;
        /** Contains if the data is sealed. */
        readonly isSealed: boolean;
        /** Contains a reference to the context (when available). */
        readonly context?: Context;
    };
    export {};
}

declare module "@tripetto/vfsm/lib/factory" {
    /** Dependencies */
    import { StrictItem, StrictMatrix } from "@marblecore/matrix";
    import { THooks } from "@marblecore/hookup";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    export interface IFactory<
        Dispatcher extends StrictMatrix<
            Dispatcher,
            Factory<Dispatcher, Ontology, Hooks>,
            void,
            Ontology
        >,
        T extends Factory<Dispatcher, Ontology, Hooks>,
        Ontology extends IOntologyProperties,
        Hooks extends THooks = void
    > {
        readonly property: string;
        new (pDispatcher: Dispatcher, pOntology: Ontology): T;
    }
    export abstract class Factory<
        Dispatcher extends StrictMatrix<
            Dispatcher,
            Factory<Dispatcher, Ontology, Hooks>,
            void,
            Ontology
        >,
        Ontology extends IOntologyProperties,
        Hooks extends THooks = void
    > extends StrictItem<Dispatcher, Hooks, Ontology> {
        /** Contains the ontology property. */
        static readonly property: string;
        /**
         * Constructs a new item.
         * @param pDispatcher Reference to the dispatcher.
         * @param sHash Specifies the unique hash for the item.
         */
        constructor(pDispatcher: Dispatcher, sHash: string);
        /** Retrieves the ancestor state. */
        get ancestor(): State | undefined;
        /** Retrieves the parent branch. */
        get branch(): Branch | undefined;
        /** Retrieves the identifier. */
        get id(): string;
        /**
         * Prepares the instance.
         * @param pAncestor Reference to the ancestor.
         * @param pBranch Reference to the branch.
         */
        prepare(
            pAncestor: State,
            pBranch?: Branch | undefined
        ): Factory<Dispatcher, Ontology, Hooks>;
        /**
         * Sets the successor state.
         * @param sMode Specifies the mode.
         * @param pState Specifies the state to set.
         */
        setSuccessor(
            sMode: "live" | "preview",
            pState: State | undefined
        ): void;
        /**
         * Retrieves a reference to the successor state.
         * @param sMode Specifies the mode.
         * @return Returns the successor state.
         */
        successor(sMode: "live" | "preview"): State | undefined;
    }
}

declare module "@tripetto/vfsm/lib/branches/culling" {
    export type TBranchCulling = "first" | "each" | "all";
}

declare module "@tripetto/vfsm/lib/states/main" {
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { States } from "@tripetto/vfsm/lib/states/states";
    export class Main extends State {
        /**
         * Constructs a new main state.
         * @param pStates Reference to the state dispatcher.
         */
        constructor(pStates: States);
        /** Enter condition for the main state. */
        onEnterCondition(): boolean;
        /** Main state is entered. */
        onEnter(): void;
        /** Leave condition for the main state. */
        onLeaveCondition(): boolean;
        /** Main state is validated. */
        onValidate(): boolean;
    }
}

declare module "@tripetto/vfsm/lib/instance/moments/snapshot" {
    /** Dependencies */
    import { IStackSnapshot } from "@tripetto/vfsm/lib/instance/stack/snapshot";
    export interface IMomentSnapshot {
        /** Contains the stack snapshot. */
        readonly a: IStackSnapshot[];
        /** Contains the stack pointer. */
        readonly b?: string;
    }
}

declare module "@tripetto/vfsm/lib/instance/orchestrator" {
    /** Dependencies */
    import { Namespace } from "@marblecore/l10n";
    import { Slot } from "@tripetto/slots";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { IChainLookup } from "@tripetto/vfsm/lib/chain";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { States } from "@tripetto/vfsm/lib/states/states";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { IDataChangeEvent } from "@tripetto/vfsm/lib/instance/data/events";
    export interface IInstanceOrchestrator {
        /** Contains the unique ontology fingerprint. */
        readonly fingerprint: string;
        /** Reference to the l10n namespace. */
        readonly l10n: Namespace;
        /** Reference to the ontology states. */
        readonly states: States;
        /** Specifies if the automatic push-forward is enabled. */
        readonly autoPushForward: boolean;
        /** Contains the preview mode. */
        readonly preview: "disabled" | "compact" | "enabled";
        /** Chain. */
        readonly chain: IChainLookup;
        /** Contains the events. */
        readonly events: {
            /** Invoked when an instance is created. */
            onInstanceCreate: (pInstance: Instance) => void;
            /** Invoked when an instance is started. */
            onInstanceStart: (pInstance: Instance) => void;
            /** Invoked when the data of an instance has changed. */
            onInstanceData: (
                pInstance: Instance,
                pDataChangeEvent: IDataChangeEvent
            ) => void;
            /** Invoked when the instance is processing. */
            onInstanceProcessing: (
                pInstance: Instance,
                bProcessing: boolean
            ) => void;
            /** Invoked when the instance wants to step forward. */
            onInstanceStepForward: (
                pInstance: Instance,
                pState: State
            ) => boolean;
            /** Invoked when the instance is updated. */
            onInstanceUpdate: (pInstance: Instance) => void;
            /** Invoked when an instance is about to end. */
            onInstanceFinishing: (
                pInstance: Instance,
                fnFinish: (bFinish: boolean, sReference?: string) => void
            ) => boolean;
            /** Invoked when an instance has finished. */
            onInstanceFinish: (pInstance: Instance) => void;
            /** Invoked when an instance is paused. */
            onInstancePause: (pInstance: Instance) => void;
            /** Invoked when an instance is stopped. */
            onInstanceStop: (pInstance: Instance) => void;
            /** Invoked when an instance ends. */
            onInstanceEnd: (
                pInstance: Instance,
                sType: "finished" | "stopped" | "paused"
            ) => void;
        };
        /** Spawns a new instance. */
        readonly spawn: (pInstance: Instance, nIndex: number) => Instance;
        /** Retrieves a dependency tree of transducers that are affected by the slot. */
        readonly dependencies: (pSlot: Slot) => {
            readonly [transducer: string]: boolean | undefined;
        };
        /** Iterates through all possible transducers and their context. */
        readonly iterate: (
            fnTransducer: (
                pTransducer: Transducer,
                pContext: Condition[],
                sContext: string
            ) => void
        ) => void;
        /** Contains the exportables stencil. */
        readonly stencil: (sType: "exportables" | "actionables") => string;
        /** Retrieves if the transducer with the specified id has changed. */
        readonly hasChanged: (sId: string, bReset: boolean) => boolean;
        /** Logger. */
        readonly logger?: (sMessage: string) => void;
    }
}

declare module "@tripetto/vfsm/lib/threading" {
    export type TThreading = "single" | "multi";
}

declare module "@tripetto/vfsm/lib/instance/data/snapshot" {
    /** Dependencies */
    import { TSerializeTypes } from "@marblecore/serializer";
    export interface IValueSnapshot {
        /** Contains the context. */
        readonly a: string;
        /** Contains the value. */
        readonly b?: TSerializeTypes;
        /** Contains the reference. */
        readonly c?: string;
        /** Specifies value flags. */
        readonly d?: number;
        /** Contains the UTC set time of the value. */
        readonly e?: number;
    }
    export interface IDataSnapshot {
        /** Contains the store. */
        readonly a: string;
        /** Contains the values. */
        readonly b?: IValueSnapshot[];
    }
}

declare module "@tripetto/vfsm/lib/instance/data/content" {
    export interface IContent {
        readonly value: string;
        readonly type: "string" | "text" | "markdown";
        readonly separator: string;
    }
}

declare module "@tripetto/vfsm/lib/instance/stack/stack" {
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { IStackOrchestrator } from "@tripetto/vfsm/lib/instance/stack/orchestrator";
    import { IStackSnapshot } from "@tripetto/vfsm/lib/instance/stack/snapshot";
    import { Item, Matrix } from "@marblecore/matrix";
    import { Opcode } from "@tripetto/vfsm/enum/opcodes";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { Validators } from "@tripetto/vfsm/lib/instance/validators/validators";
    export class Stack extends Item {
        /**
         * Creates a new stack.
         * @param pStack Reference to the stack matrix.
         * @param pOrchestrator Reference to the stack orchestrator.
         * @param pAncestor Reference to the stack ancestor.
         * @param pState Reference to a state.
         * @param pBranch Reference to a branch.
         * @param pCondition Reference to a condition.
         * @param pBranched Contains a reference to the active branch.
         * @param pValidators Reference to the transducer validators.
         * @param nOpcode Specifies the opcode.
         * @param nCounter Specifies the number of loops.
         * @param nSteps Specifies the number of steps.
         * @param bReturn Specifies if there is a return value.
         * @param bReturnOverride Specifies if a return override value is present.
         * @param pReturn Specifies the return state.
         * @param bBranching Specifies if the stack is branching.
         */
        static create(
            pStack: Matrix<Stack>,
            pOrchestrator: IStackOrchestrator,
            pState: State | undefined,
            pAncestor?: Stack | undefined,
            pBranch?: Branch | undefined,
            pCondition?: Condition | undefined,
            pBranched?: Branch | undefined,
            pValidators?: Validators | undefined,
            nOpcode?: Opcode,
            nCounter?: number,
            bReturn?: boolean,
            bReturnOverride?: boolean,
            pReturn?: State | undefined,
            bBranching?: boolean
        ): Stack;
        /** Retrieves the identifier. */
        get id(): string;
        /** Retrieves the instance of the stack. */
        get instance(): Instance;
        /** Contains a reference to the ancestor stack. */
        get ancestor(): Stack | undefined;
        /** Retrieves the stack trace. */
        get trace(): string;
        /** Reference to the state for the stack. */
        get state(): State | undefined;
        /** Retrieves the validated transducers for the stack. */
        get transducers(): Transducer[];
        /** Reference to the branch for the stack. */
        get branch(): Branch | undefined;
        /** Reference to the condition for the stack. */
        get condition(): Condition | undefined;
        /** Retrieves the mutable stack context. */
        get mutableContext(): Condition | undefined;
        /** Retrieves the immutable stack context. */
        get immutableContext(): Condition | undefined;
        /** Retrieves the stack checksum. */
        get checksum(): string;
        /** Retrieves if the stack is branching. */
        get isBranch(): boolean;
        /** Specifies if the stack is scheduled for the pursue step. */
        get isScheduled(): boolean;
        /** Retrieves if the stack branch is at the finish. */
        get isAtFinish(): boolean;
        /** Retrieves if the stack is evaluating. */
        get isEvaluating(): boolean;
        /**
         * Performs a step.
         * @param nOpcode Specifies the opcode to be executed.
         */
        step(nOpcode?: Opcode): void;
        /**
         * Invalidates a transducer.
         * @param pTransducer Specifies the transducer to invalidate.
         */
        invalidate(pTransducer: Transducer): void;
        /**
         * Retrieves the validation state of the supplied transducer.
         * @param pTransducer Specifies the transducer.
         * @return Returns the validation state.
         */
        validationOf(pTransducer: Transducer): "unknown" | "pass" | "fail";
        /** Cancels the stack. */
        cancel(): void;
        /** Destroys the stack. */
        destroy(): void;
        /**
         * Snapshots the stack.
         * @return Returns the stack snapshot.
         */
        snapshot(): IStackSnapshot;
    }
}

declare module "@tripetto/vfsm/lib/instance/data/proxy" {
    /** Dependencies */
    import { TSerializeTypes } from "@marblecore/serializer";
    import { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    import { IContent } from "@tripetto/vfsm/lib/instance/data/content";
    export interface IValueProxy {
        /** Contains the context of the proxy. */
        readonly context: string | undefined;
        /** Contains the value as a string. */
        readonly string: string;
        /** Contains the content for the variable. */
        readonly content: IContent[];
        /** Contains the value. */
        readonly value?: TSerializeTypes;
        /** Contains the value reference. */
        readonly reference?: string;
        /** Contains if a value is present. */
        readonly hasValue: boolean;
        /** Contains a list of slot references covered by the proxy. */
        readonly refs: string[];
        /** Updates the immutable value. */
        readonly update: (pValue: ImmutableValue) => void;
    }
}

declare module "@tripetto/vfsm/lib/constructors" {
    /** Dependencies */
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { TState } from "@tripetto/vfsm/lib/states/type";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { TTransducer } from "@tripetto/vfsm/lib/transducers/type";
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { TBranch } from "@tripetto/vfsm/lib/branches/type";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { TCondition } from "@tripetto/vfsm/lib/conditions/type";
    export interface IConstructors<
        S extends State,
        T extends Transducer,
        B extends Branch,
        C extends Condition
    > {
        readonly state: TState<S>;
        readonly transducer: TTransducer<T>;
        readonly branch: TBranch<B>;
        readonly condition: TCondition<C>;
    }
}

declare module "@tripetto/vfsm/lib/branches/type" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Branches } from "@tripetto/vfsm/lib/branches/branches";
    import { IBranch } from "@tripetto/vfsm/lib/branches/interface";
    import { IFactory } from "@tripetto/vfsm/lib/factory";
    import { TAny } from "@marblecore/functional";
    import { TBranchEvents } from "@tripetto/vfsm/lib/branches/events";
    export type TBranch<
        T extends Branch = Branch,
        Ontology extends IBranch = TAny
    > = IFactory<Branches, T, Ontology, TBranchEvents>;
}

declare module "@tripetto/vfsm/lib/conditions/type" {
    /** Dependencies */
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { Conditions } from "@tripetto/vfsm/lib/conditions/conditions";
    import { ICondition } from "@tripetto/vfsm/lib/conditions/interface";
    import { IFactory } from "@tripetto/vfsm/lib/factory";
    import { TAny } from "@marblecore/functional";
    export type TCondition<
        T extends Condition = Condition,
        Ontology extends ICondition = TAny
    > = IFactory<Conditions, T, Ontology>;
}

declare module "@tripetto/vfsm/lib/states/type" {
    /** Dependencies */
    import { IFactory } from "@tripetto/vfsm/lib/factory";
    import { IState } from "@tripetto/vfsm/lib/states/interface";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { States } from "@tripetto/vfsm/lib/states/states";
    import { TAny } from "@marblecore/functional";
    import { TStateEvents } from "@tripetto/vfsm/lib/states/events";
    export type TState<
        T extends State = State,
        Ontology extends IState = TAny
    > = IFactory<States, T, Ontology, TStateEvents>;
}

declare module "@tripetto/vfsm/lib/transducers/type" {
    /** Dependencies */
    import { IFactory } from "@tripetto/vfsm/lib/factory";
    import { ITransducer } from "@tripetto/vfsm/lib/transducers/interface";
    import { TAny } from "@marblecore/functional";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { Transducers } from "@tripetto/vfsm/lib/transducers/transducers";
    import { TTransducerEvents } from "@tripetto/vfsm/lib/transducers/events";
    export type TTransducer<
        T extends Transducer = Transducer,
        Ontology extends ITransducer = TAny
    > = IFactory<Transducers, T, Ontology, TTransducerEvents>;
}

declare module "@tripetto/vfsm/lib/instance/stack/snapshot" {
    /** Dependencies */
    import { IValidatorSnapshot } from "@tripetto/vfsm/lib/instance/validators/snapshot";
    import { Opcode } from "@tripetto/vfsm/enum/opcodes";
    export interface IStackSnapshot {
        /** Contains the stack. */
        readonly a: string;
        /** Contains the counter. */
        readonly b?: number;
        /** Contains the opcode. */
        readonly c?: Opcode;
        /** Contains the flags. */
        readonly d?: number;
        /** Contains the ancestor. */
        readonly e?: string;
        /** Contains the state. */
        readonly f?: string;
        /** Contains the branch. */
        readonly g?: string;
        /** Contains the active branch. */
        readonly h?: string;
        /** Contains the condition. */
        readonly i?: string;
        /** Contains the return state. */
        readonly j?: string;
        /** Contains the transducer snapshot. */
        readonly k?: IValidatorSnapshot[];
    }
}

declare module "@tripetto/vfsm/lib/instance/stack/orchestrator" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { IChainLookup } from "@tripetto/vfsm/lib/chain";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { Stack } from "@tripetto/vfsm/lib/instance/stack/stack";
    import { State } from "@tripetto/vfsm/lib/states/state";
    export interface IStackOrchestrator {
        /** Reference to the instance. */
        readonly instance: Instance;
        /** Contains if the preview mode is enabled. */
        readonly preview: boolean;
        /** Contains if the compact preview mode is enabled. */
        readonly previewCompact: boolean;
        /** Spawns a new instance. */
        readonly spawn: (pState: State) => void;
        /** Sets the validation state of the instance. */
        readonly validated: (pState: State) => void;
        /** Invoked when the stack makes a cycle. */
        readonly cycle: (bCycle: boolean) => void;
        /** Invoked when a stack is canceled. */
        readonly cancel: () => void;
        /** Snapshots. */
        readonly snapshot: {
            /** Pushes the current moment snapshot. */
            readonly push: () => void;
            /** Makes a snapshot of the current stack. */
            readonly make: () => void;
        };
        /** Pointer. */
        readonly pointer: {
            /** Retrieves the stack pointer. */
            readonly get: () => Stack | undefined;
            /** Assures a stack is bind to the pointer. */
            readonly assure: (pStack: Stack) => Stack | undefined;
            /** Releases the current pointer. */
            readonly release: () => void;
        };
        /** Chain. */
        readonly chain: IChainLookup;
        /** Logger. */
        readonly logger: (
            pStack: Stack,
            pAncestor: Stack | undefined,
            pState: State | undefined,
            pBranch: Branch | undefined,
            pCondition: Condition | undefined,
            pReturn: State | true | undefined
        ) => void;
    }
}

declare module "@tripetto/vfsm/enum/opcodes" {
    export const enum Opcode {
        /** An item is pushed to the stack. */
        Push = 0,
        /** An item is awaiting the outcome of the branch conditions. */
        Branch = 1,
        /**
         * An item is awaiting further processing after the branch conditions
         * are processed.
         */
        BranchWait = 2,
        /** An item state is entered. */
        Enter = 3,
        /** The enter conditions for an item are processed. */
        EnterCondition = 4,
        /** Awaiting the enter conditions. */
        EnterConditionWait = 5,
        /** Waiting for state feedback. */
        Wait = 6,
        /** The leave conditions for an item are processed. */
        LeaveCondition = 7,
        /** Awaiting the leave conditions. */
        LeaveConditionWait = 8,
        /** An item is awaiting pursuing actions. */
        Pursue = 9,
        /** An item is popped. */
        Pop = 10,
    }
}

declare module "@tripetto/vfsm/lib/instance/validators/validators" {
    /** Dependencies */
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { IValidatorSnapshot } from "@tripetto/vfsm/lib/instance/validators/snapshot";
    import { Matrix } from "@marblecore/matrix";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { Validator } from "@tripetto/vfsm/lib/instance/validators/validator";
    export class Validators extends Matrix<Validator> {
        /** Retrieves the validated transducers. */
        get transducers(): Transducer[];
        /** Retrieves if one or more validators are enabled. */
        get isEnabled(): boolean;
        /** Retrieves if all validators passed the validation. */
        get isPassed(): boolean;
        /**
         * Creates a new transducer validator.
         * @param pTransducer Reference to the transducer.
         * @param pContext Context for the validator.
         * @param bEnabled Specifies if the transducer is enabled.
         */
        create(
            pTransducer: Transducer,
            pContext: Context,
            bEnabled: boolean
        ): Validator;
        /**
         * Resumes a validator.
         * @param pTransducer Reference to the transducer.
         * @param pContext Context for the validator.
         * @param pSnapshot Reference to the snapshot data.
         */
        resume(
            pTransducer: Transducer,
            pContext: Context,
            pSnapshot: IValidatorSnapshot
        ): Validator;
        /**
         * Retrieves the validator for the specified transducer.
         * @param pTransducer Specifies the transducer.
         * @return Returns the validator instance or `undefined`.
         */
        forTransducer(pTransducer: Transducer): Validator | undefined;
        /** Clears all validators. */
        clear(): void;
    }
}

declare module "@tripetto/vfsm/lib/instance/validators/snapshot" {
    export interface IValidatorSnapshot {
        /** Contains the transducer. */
        readonly a: string;
        /** Contains if the validator flags. */
        readonly b: number;
    }
}

declare module "@tripetto/vfsm/lib/instance/validators/validator" {
    /** Dependencies */
    import { ICallbackPromise } from "@marblecore/callback";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { Item } from "@marblecore/matrix";
    import { IValidatorSnapshot } from "@tripetto/vfsm/lib/instance/validators/snapshot";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    export class Validator extends Item {
        /**
         * Constructs a new transducer validator.
         * @param pTransducer Reference to the transducer.
         * @param pContext Context of the validator.
         * @param bEnabled Specifies if the transducer is enabled.
         * @param bPass Specifies if the validation is passed.
         */
        constructor(
            pTransducer: Transducer,
            pContext: Context,
            bEnabled: boolean,
            bPass?: boolean
        );
        /** Retrieves a reference to the transducer. */
        get transducer(): Transducer;
        /** Specifies if the transducer is enabled. */
        get isEnabled(): boolean;
        /** Retrieves if the transducer validation is passed. */
        get isPassed(): boolean;
        /** Retrieves if the transducer validation failed. */
        get isFailed(): boolean;
        /**
         * Validates the transducer.
         * @param pState Reference to the state.
         * @param sType Specifies the type of validation.
         * @param bPreview Specifies if the preview mode is enabled.
         * @param fnAssert Specifies the stack assert function.
         * @param fnDone Specifies the callback function that is invoked when the
         * validation is completed.
         * @return Returns `true` if the transducer was validated.
         */
        validate(
            pState: State | undefined,
            sType: "initial" | "revalidate",
            bPreview: boolean,
            fnAssert: (pPromise: ICallbackPromise<boolean>) => boolean,
            fnDone: () => void
        ): boolean;
        /** Invalidates the validation state. */
        invalidate(): void;
        /** Cancels a pending validation. */
        cancel(): void;
        /**
         * Snapshots the stack transducer validator.
         * @return Returns the stack transducer snapshot.
         */
        snapshot(): IValidatorSnapshot | undefined;
    }
}
